@@ -4,6 +4,7 @@ Simple Apereo Central Authentication Service (CAS) client
 import logging as _logging
 
 from .core import (
+    AsyncCASClient,
     CASClient,
     CASError,
     CASInvalidServiceError,
@@ -11,9 +12,10 @@ from .core import (
     CASUser,
 )
 
-__version__ = "0.0.6"
+__version__ = "0.0.7"
 
 __all__ = [
+    "AsyncCASClient",
     "CASClient",
     "CASUser",
     "CASError",
