@@ -1,19 +1,23 @@
 """
 Central Authentication Service (CAS) client
 """
+import asyncio
 import dataclasses
 import logging
 import urllib.parse
 import urllib.request
 import xml.etree.ElementTree
+from collections.abc import Awaitable, Callable
 from typing import Optional, Union, cast
 
 logger = logging.getLogger(__name__)
 
+CAS_NS = {"cas": "http://www.yale.edu/tp/cas"}
+CAS_VALIDATE_ENCODING = "utf-8"
+CAS_VALIDATE_TIMEOUT = 10.0
 
-def _fetch_url(url: str, timeout: float = 10.0) -> bytes:
-    with urllib.request.urlopen(url, timeout=timeout) as f:
-        return cast(bytes, f.read())
+HTTPGetFunc = Callable[[str, float], str]
+AsyncHTTPGetFunc = Callable[[str, float], Awaitable[str]]
 
 
 class CASError(Exception):
@@ -23,13 +27,13 @@ class CASError(Exception):
 
 
 class CASInvalidServiceError(CASError):
-    def __init__(self, error_code: str, *args: Optional[str]) -> None:
-        super().__init__(error_code, *args)
+    def __init__(self, *args: Optional[str]) -> None:
+        super().__init__("INVALID_SERVICE", *args)
 
 
 class CASInvalidTicketError(CASError):
-    def __init__(self, error_code: str, *args: Optional[str]) -> None:
-        super().__init__(error_code, *args)
+    def __init__(self, *args: Optional[str]) -> None:
+        super().__init__("INVALID_TICKET", *args)
 
 
 @dataclasses.dataclass
@@ -41,57 +45,12 @@ class CASUser:
         return dataclasses.asdict(self)
 
 
-class CASClient:
-    CAS_NS = {"cas": "http://www.yale.edu/tp/cas"}
-    CAS_VALIDATE_ENCODING = "utf-8"
-    CAS_VALIDATE_TIMEOUT = 10.0
-
-    def __init__(
-        self,
-        login_url: str,
-        logout_url: str,
-        validate_url: str,
-    ) -> None:
+class BaseCASClient:
+    def __init__(self, login_url: str, logout_url: str, validate_url: str) -> None:
         self.login_url = login_url
         self.logout_url = logout_url
         self.validate_url = validate_url
 
-    @classmethod
-    def from_base_url(
-        cls,
-        base_url: str,
-        *,
-        login_path: str = "/login",
-        logout_path: str = "/logout",
-        validate_path: str = "/p3/serviceValidate",
-    ) -> "CASClient":
-        return cls(
-            login_url=urllib.parse.urljoin(base_url, login_path),
-            logout_url=urllib.parse.urljoin(base_url, logout_path),
-            validate_url=urllib.parse.urljoin(base_url, validate_path),
-        )
-
-    def validate(
-        self,
-        service_url: str,
-        ticket: str,
-        *,
-        timeout: Optional[float] = None,
-        **kwargs: str,
-    ) -> CASUser:
-        if timeout is None:
-            timeout = self.CAS_VALIDATE_TIMEOUT
-        target_validate = self.build_validate_url(service_url, ticket, **kwargs)
-        logger.debug("Validating %s", target_validate)
-        try:
-            resp_data = _fetch_url(target_validate, timeout=timeout)
-            resp_text = resp_data.decode(self.CAS_VALIDATE_ENCODING)
-        except Exception as exc:
-            raise CASError(repr(exc)) from exc
-        else:
-            logger.debug("Response:\n%s", resp_text)
-            return self.parse_cas_response(resp_text)
-
     def build_login_url(
         self,
         service: str,
@@ -120,58 +79,171 @@ class CASClient:
         qs = urllib.parse.urlencode(params)
         return f"{self.validate_url}?{qs}"
 
-    def parse_cas_response(self, cas_response: str) -> CASUser:
-        try:
-            root = xml.etree.ElementTree.fromstring(cas_response)
-        except Exception as exc:
-            raise CASError("INVALID_RESPONSE", repr(exc)) from exc
-        else:
-            return self.parse_cas_xml(root)
-
-    def parse_cas_xml(self, root: xml.etree.ElementTree.Element) -> CASUser:
-        user_elem = root.find("cas:authenticationSuccess/cas:user", self.CAS_NS)
-        if user_elem is not None:
-            attr_elem = root.find(
-                "cas:authenticationSuccess/cas:attributes", self.CAS_NS
-            )
-            return self.parse_cas_xml_user(user_elem, attr_elem)
-        raise self.parse_cas_xml_error(root)
-
-    def parse_cas_xml_user(
-        self,
-        user_elem: xml.etree.ElementTree.Element,
-        attr_elem: Optional[xml.etree.ElementTree.Element],
-    ) -> CASUser:
-        if user_elem.text is None:
-            raise CASError("USERNAME_NOT_IN_RESPONSE")
-        cas_user = CASUser(userid=user_elem.text)
-        if attr_elem is not None:
-            tag_ns = "{" + self.CAS_NS["cas"] + "}"
-            for e in attr_elem:
-                attr_name = e.tag.replace(tag_ns, "", 1)
-                cas_user.attributes[attr_name] = e.text
-        return cas_user
-
-    def parse_cas_xml_error(
-        self,
-        root: xml.etree.ElementTree.Element,
-    ) -> CASError:
-        error_code = "Unknown"
-        error_elem = root.find("cas:authenticationFailure", self.CAS_NS)
-        if error_elem is not None:
-            error_code = error_elem.attrib.get("code", error_code)
-            error_text = error_elem.text
-            if error_code == "INVALID_TICKET":
-                return CASInvalidTicketError(error_code, error_text)
-            if error_code == "INVALID_SERVICE":
-                return CASInvalidServiceError(error_code, error_text)
-        return CASError(error_code)
-
     def __repr__(self) -> str:
         return (
-            "CASClient("
+            f"{self.__class__.__name__}("
             f"login_url={self.login_url!r}, "
             f"logout_url={self.logout_url!r}, "
             f"validate_url={self.validate_url!r}"
             ")"
         )
+
+
+class CASClient(BaseCASClient):
+    def __init__(
+        self,
+        login_url: str,
+        logout_url: str,
+        validate_url: str,
+        http_get_func: Optional[HTTPGetFunc] = None,
+    ) -> None:
+        super().__init__(login_url, logout_url, validate_url)
+        if http_get_func is None:
+            self._http_get = cast(HTTPGetFunc, _http_get)
+        else:
+            self._http_get = http_get_func
+
+    @classmethod
+    def from_base_url(
+        cls,
+        base_url: str,
+        *,
+        login_path: str = "/login",
+        logout_path: str = "/logout",
+        validate_path: str = "/p3/serviceValidate",
+        http_get_func: Optional[HTTPGetFunc] = None,
+    ) -> "CASClient":
+        return cls(
+            login_url=urllib.parse.urljoin(base_url, login_path),
+            logout_url=urllib.parse.urljoin(base_url, logout_path),
+            validate_url=urllib.parse.urljoin(base_url, validate_path),
+            http_get_func=http_get_func,
+        )
+
+    def validate(
+        self,
+        service_url: str,
+        ticket: str,
+        *,
+        timeout: Optional[float] = None,
+        **kwargs: str,
+    ) -> CASUser:
+        if timeout is None:
+            timeout = CAS_VALIDATE_TIMEOUT
+        target_validate = self.build_validate_url(service_url, ticket, **kwargs)
+        logger.debug("Validating %s", target_validate)
+        try:
+            resp_text = self._http_get(target_validate, timeout)
+        except Exception as exc:
+            raise CASError(repr(exc)) from exc
+        else:
+            logger.debug("Response:\n%s", resp_text)
+            return parse_cas_response(resp_text)
+
+
+class AsyncCASClient(BaseCASClient):
+    def __init__(
+        self,
+        login_url: str,
+        logout_url: str,
+        validate_url: str,
+        http_get_func: Optional[AsyncHTTPGetFunc] = None,
+    ) -> None:
+        super().__init__(login_url, logout_url, validate_url)
+        if http_get_func is None:
+            self._http_get = cast(AsyncHTTPGetFunc, _async_http_get)
+        else:
+            self._http_get = http_get_func
+
+    @classmethod
+    def from_base_url(
+        cls,
+        base_url: str,
+        *,
+        login_path: str = "/login",
+        logout_path: str = "/logout",
+        validate_path: str = "/p3/serviceValidate",
+        http_get_func: Optional[AsyncHTTPGetFunc] = None,
+    ) -> "AsyncCASClient":
+        return cls(
+            login_url=urllib.parse.urljoin(base_url, login_path),
+            logout_url=urllib.parse.urljoin(base_url, logout_path),
+            validate_url=urllib.parse.urljoin(base_url, validate_path),
+            http_get_func=http_get_func,
+        )
+
+    async def validate(
+        self,
+        service_url: str,
+        ticket: str,
+        *,
+        timeout: Optional[float] = None,
+        **kwargs: str,
+    ) -> CASUser:
+        if timeout is None:
+            timeout = CAS_VALIDATE_TIMEOUT
+        target_validate = self.build_validate_url(service_url, ticket, **kwargs)
+        logger.debug("Validating %s", target_validate)
+        try:
+            resp_text = await self._http_get(target_validate, timeout)
+        except Exception as exc:
+            raise CASError(repr(exc)) from exc
+        else:
+            logger.debug("Response:\n%s", resp_text)
+            return parse_cas_response(resp_text)
+
+
+def parse_cas_response(cas_response: str) -> CASUser:
+    try:
+        root = xml.etree.ElementTree.fromstring(cas_response)
+    except Exception as exc:
+        raise CASError("INVALID_RESPONSE", repr(exc)) from exc
+    else:
+        return parse_cas_xml(root)
+
+
+def parse_cas_xml(root: xml.etree.ElementTree.Element) -> CASUser:
+    user_elem = root.find("cas:authenticationSuccess/cas:user", CAS_NS)
+    if user_elem is not None:
+        attr_elem = root.find("cas:authenticationSuccess/cas:attributes", CAS_NS)
+        return parse_cas_xml_user(user_elem, attr_elem)
+    raise parse_cas_xml_error(root)
+
+
+def parse_cas_xml_user(
+    user_elem: xml.etree.ElementTree.Element,
+    attr_elem: Optional[xml.etree.ElementTree.Element],
+) -> CASUser:
+    if user_elem.text is None:
+        raise CASError("USERNAME_NOT_IN_RESPONSE")
+    cas_user = CASUser(userid=user_elem.text)
+    if attr_elem is not None:
+        tag_ns = "{" + CAS_NS["cas"] + "}"
+        for e in attr_elem:
+            attr_name = e.tag.replace(tag_ns, "", 1)
+            cas_user.attributes[attr_name] = e.text
+    return cas_user
+
+
+def parse_cas_xml_error(root: xml.etree.ElementTree.Element) -> CASError:
+    error_code = "Unknown"
+    error_elem = root.find("cas:authenticationFailure", CAS_NS)
+    if error_elem is not None:
+        error_code = error_elem.attrib.get("code", error_code)
+        error_text = error_elem.text
+        if error_code == "INVALID_TICKET":
+            return CASInvalidTicketError(error_text)
+        if error_code == "INVALID_SERVICE":
+            return CASInvalidServiceError(error_text)
+    return CASError(error_code)
+
+
+def _http_get(url: str, timeout: float = 10.0) -> str:
+    with urllib.request.urlopen(url, timeout=timeout) as f:
+        data = cast(bytes, f.read())
+        return data.decode(CAS_VALIDATE_ENCODING)
+
+
+async def _async_http_get(url: str, timeout: float = 10.0) -> str:
+    fut = asyncio.to_thread(_http_get, url, timeout)
+    return await asyncio.wait_for(fut, timeout)
